#define __KERNEL__		/* kernel code */ 
#define MODULE			/* always as a module */ 

#include <linux/module.h>
#include <linux/version.h>

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>

#include <linux/fs.h>		/* for file_operations */
#include <linux/pci.h>      	/* for pci_find_device */
#include <linux/slab.h>      	/* for kmalloc */
#include <asm/segment.h>	/* for memcpy */
#include <asm/io.h>		/* for virt_to_phys */
#include <linux/byteorder/little_endian.h>

#include "debug.h"
#include "3c905c.h"

EXPORT_NO_SYMBOLS; 

MODULE_AUTHOR("FBarreto & Guto");
MODULE_DESCRIPTION("3COM 905c (Tornado) Driver");

/* MODULE PARAMETERS */
int dev = C905C_MAJOR;
MODULE_PARM(dev,"i");
MODULE_PARM_DESC(dev,"Major device number");

/* PROTOTYPES */
int init_module(void); 
void cleanup_module(void);
static int c905c_open(struct inode *inode, struct file *filp);
static int c905c_close(struct file *filp);
static ssize_t c905c_read(struct file *, char *, size_t, loff_t *);
static ssize_t c905c_write(struct file *, const char *, size_t, loff_t *);

/* DMA BUFFER FILE OPERATIONS */
static struct file_operations c905c_fops = {
    owner:              NULL,
    llseek:             NULL,
    read:               c905c_read,
    write:              c905c_write,
    readdir:            NULL,
    poll:               NULL,
    ioctl:              NULL,
    mmap:               NULL,
    open:               c905c_open,
    flush:              c905c_close,
    release:            NULL,
    fsync:              NULL,
    fasync:             NULL,
    lock:               NULL,
    readv:              NULL,
    writev:             NULL,
    sendpage:           NULL,
    get_unmapped_area:  NULL
};

static struct C905C_NIC * nic;

int init_module(void)
{
    int err;
    struct pci_dev *pci_dev;
    static struct PCI_Device pcidev;
    
    debug(DB_TRACE, "Loading 3C905c Driver (dev=%d)\n", dev);

    /* Look for our PCI device */
    pci_dev = pci_find_device(PCI_VENDOR_ID_3COM, PCI_DEVICE_ID_3COM_905C, 0);
    if(!pci_dev)
	return -ENODEV;

    /* (Re)configure device for bus mastering */
    pci_read_config_word(pci_dev, PCI_COMMAND, &pcidev.cmd);
    if(!(pcidev.cmd & PCI_COMMAND_MEMORY)) {
	pcidev.cmd |= PCI_COMMAND_MEMORY;
	debug(DB_WARNING, "PCI_COMMAND_MEMORY not set for device!\n");
    }
    if(!(pcidev.cmd & PCI_COMMAND_MASTER)) {
	pcidev.cmd |= PCI_COMMAND_MASTER;
	debug(DB_WARNING, "PCI_COMMAND_MASTER not set for device!\n");
    }
    if(!(pcidev.cmd & PCI_COMMAND_INVALIDATE)) {
	pcidev.cmd |= PCI_COMMAND_INVALIDATE;
	debug(DB_WARNING, "PCI_COMMAND_INVALIDATE not set for device!\n");
    }
    pci_write_config_word(pci_dev, PCI_COMMAND, pcidev.cmd);

    /* Get device's configuration */
    pci_read_config_word(pci_dev, PCI_COMMAND, &pcidev.cmd);
    pci_read_config_word(pci_dev, PCI_STATUS, &pcidev.stat);
    pci_read_config_byte(pci_dev, PCI_REVISION_ID, &pcidev.rev);

    /* Get reagion 1 info */
    pci_read_config_dword(pci_dev, PCI_BASE_ADDRESS_1,
			  &pcidev.reg[1].phy_addr);
    cli();
    pci_write_config_dword(pci_dev, PCI_BASE_ADDRESS_1, ~0);
    pci_read_config_dword(pci_dev, PCI_BASE_ADDRESS_1, &pcidev.reg[1].size);
    pci_write_config_dword(pci_dev, PCI_BASE_ADDRESS_1,
			   pcidev.reg[1].phy_addr);
    sti();

    /* Check for memory region 	*/
    if(!pcidev.reg[1].size 
       || (pcidev.reg[1].phy_addr & PCI_BASE_ADDRESS_SPACE_IO))
	return -ENODEV;
	
    /* We only handle 32 bit addressed memory above 1 Mb */
    if((pcidev.reg[1].phy_addr & PCI_BASE_ADDRESS_MEM_TYPE_MASK) &
       (PCI_BASE_ADDRESS_MEM_TYPE_1M | PCI_BASE_ADDRESS_MEM_TYPE_64)) 
	return -ENODEV;
	
    /* Adjust address and size */
    pcidev.reg[1].phy_addr &= PCI_BASE_ADDRESS_MEM_MASK;
    pcidev.reg[1].size = ~(pcidev.reg[1].size & PCI_BASE_ADDRESS_MEM_MASK) + 1;
	
    /* Map ragion into kernel address space */
    pcidev.reg[1].log_addr = (unsigned long)
	ioremap_nocache(pcidev.reg[1].phy_addr, pcidev.reg[1].size); 
    if(!pcidev.reg[1].log_addr)
	return -ENODEV;

    debug(DB_INFO, "pcidev->ctrl=%x\n", pcidev.cmd);
    debug(DB_INFO, "pcidev->stat=%x\n", pcidev.stat);
    debug(DB_INFO, "pcidev->rev=%x\n", pcidev.rev);
    debug(DB_INFO, "pcidev->regs=%x\n", pcidev.regs);
    debug(DB_INFO, "pcidev->reg[1]->phy_addr=%x\n", pcidev.reg[1].phy_addr);
    debug(DB_INFO, "pcidev->reg[1]->size=%d\n", pcidev.reg[1].size);
    
    nic = (struct C905C_NIC *)pcidev.reg[1].log_addr;

    // Enable nic
    if(nic->win.w1.IntStatus & ( 1 << 12 )) {
	debug(DB_WARNING, "3C905c nic busy!\n");
	while (nic->win.w1.IntStatus & ( 1 << 12 ));
    } else {
	nic->win.w1.IntStatus |= (9 << 11); // Tx enable
	nic->win.w1.IntStatus |= (4 << 11); // Rx enable
    } 		
       
    SET_MODULE_OWNER (&c905c_fops);

    /* Register the device driver */
    err = register_chrdev(dev, "3C905c", &c905c_fops);
    if(err < 0)
	return err;
    
    return 0;
}

void cleanup_module(void)
{
    debug(DB_TRACE, "Unloading 3C905c\n");

    /* Unregister the device driver */
    unregister_chrdev(dev, "3C905c");
}

int c905c_open(struct inode *inode, struct file *filp)
{
    int unit;

    unit = MINOR(inode->i_rdev);

    debug(DB_TRACE, "3C905c_open(unit=%d)\n", unit);

    return 0;
}

int c905c_close(struct file *filp)
{
    int unit;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev);

    debug(DB_TRACE, "3C905c_close(unit=%d)\n", unit);

    return 0;
}

static ssize_t c905c_write(struct file * f, const char * b, size_t l,
			    loff_t * o)
{
    struct Frame * frame;
    struct DPD * dpd;

    debug(DB_TRACE, "3C905c_write(fil=%p,buf=%p,len=%d,off=%p)\n",
	  f, b, l, o);

    frame = kmalloc(sizeof(struct Frame), GFP_DMA);
    if(!frame)
	return -ENOMEM;

    frame->header.dst[0]=0xff;
    frame->header.dst[1]=0xff;
    frame->header.dst[2]=0xff;
    frame->header.dst[3]=0xff;
    frame->header.dst[4]=0xff;
    frame->header.dst[5]=0xff;
    
    frame->header.src[0]=0x00;
    frame->header.src[1]=0x01;
    frame->header.src[2]=0x02;
    frame->header.src[3]=0x03;
    frame->header.src[4]=0x04;
    frame->header.src[5]=0x05;	

    frame->header.type=htons(0x0801);

    //copy_from_user(frame,arg,sizeof(struct frame));

    strcpy(frame->payload, "3C905c test!");

    dpd = kmalloc(sizeof(struct DPD),GFP_DMA);
    if(!dpd) {
	kfree(frame);
	return -ENOMEM;
    }

    dpd->DnFragAddress = virt_to_phys(frame);
    dpd->DnFragLen = (1 << 31) + 1500;
    dpd->FrameStartHeader=0;
    dpd->DnNextPtr = 0;

    nic->DnListPtr = virt_to_phys(dpd);

    debug(DB_INFO, "dnComplete: %u\n", dpd->FrameStartHeader);

    kfree(frame);
    kfree(dpd);

    return l;
}

static ssize_t c905c_read(struct file * f, char * b, size_t l, loff_t * o)
{
    struct Frame * frame;
    struct UPD * upd;
    u_short 	status;
    int	i;

    frame = kmalloc(sizeof(struct Frame), GFP_DMA);
    if(!frame)
	return -ENOMEM;
		    
    upd = kmalloc(sizeof(struct UPD),GFP_DMA);
    if(!upd) {
	kfree(frame);
	return -ENOMEM;
    }

    upd->UpFragAddress = virt_to_phys(frame);
    upd->UpFragLen = (1 << 31) + 1500;   
    upd->UpPktStatus = 0;
    upd->UpNextPtr = 0;

    nic->UpListPtr = virt_to_phys(upd);

    while(!upd->UpPktStatus) {
	if (current->need_resched)
	    schedule();
	if (signal_pending(current)) 
	    break;
    };

    debug(DB_INFO, "%X:%X:%X:%X:%X:%X - %X:%X:%X:%X:%X:%X - %hu\n",
	  frame->header.dst[0], frame->header.dst[1], frame->header.dst[2],
	  frame->header.dst[3], frame->header.dst[4], frame->header.dst[5],
	  frame->header.src[0], frame->header.src[1], frame->header.src[2],
	  frame->header.src[3], frame->header.src[4], frame->header.src[5],
	  ntohs(frame->header.type)); 

    status = (upd->UpPktStatus & 8191);
	
    debug(DB_INFO, "Data size: %d\n Data: ",status);
    for(i=0;i<status-14;i++) debug(DB_INFO,"%c",frame->payload[i]);

    kfree(frame);
    kfree(upd);
    
    return 0;
}
